import os

# Default to None for PROJECT_ROOT
PROJECT_ROOT = None

def find_git_repo_with_config(start_dir=None):
    """
    Searches upwards from the specified directory to find the first Git repository
    containing a 'config' directory. Raises FileNotFoundError if no such repository is found.
    """
    if start_dir is None:
        start_dir = os.getcwd()  # Default to the current working directory

    current_dir = start_dir
    while current_dir != os.path.dirname(current_dir):  # Stop when reaching the root directory
        git_dir = os.path.join(current_dir, '.git')
        config_dir = os.path.join(current_dir, 'config')
        
        # Check if both .git and 'config' directories exist
        if os.path.isdir(git_dir) and os.path.isdir(config_dir):
            return current_dir
        
        current_dir = os.path.dirname(current_dir)  # Move up one level

    # Raise error if no matching repository is found
    raise FileNotFoundError("No project root found (git repository containing the 'config' directory).")

# Try to set PROJECT_ROOT automatically
try:
    PROJECT_ROOT = find_git_repo_with_config()
    print(f"PROJECT_ROOT is {PROJECT_ROOT}")

except FileNotFoundError as e:
    # Handle the case when automatic discovery fails
    print(e)

    # Prompt the user to manually enter the project root directory
    PROJECT_ROOT = input("Enter the path to the project root (where .git and config directories are located): ").strip()

    # Check if the user entered a valid path
    if os.path.isdir(PROJECT_ROOT) and os.path.isdir(os.path.join(PROJECT_ROOT, '.git')) and os.path.isdir(os.path.join(PROJECT_ROOT, 'config')):
        print(f"PROJECT_ROOT is set to: {PROJECT_ROOT}")
    else:
        print(f"The directory you entered is not valid. Please ensure it contains a .git directory and a 'config' directory.")
        # You could raise an error or re-prompt the user here
        raise ValueError(f"Invalid project root: {PROJECT_ROOT}")

# Import package components after PROJECT_ROOT is set
from .experiment import Experiment
from .blend import Blend
from .configuration import Config
from .hydration import Hydration

# Now you can safely create your config
config = Config()



import matplotlib as mpl
import matplotlib.pyplot as plt

def set_plot_theme():
    """Set plot theme"""

    params = {
        "axes.labelsize": 7,
        "axes.titlesize": 6,
        "xtick.labelsize": 6,
        "ytick.labelsize": 6,
        "axes.titlepad": 1,
        "axes.labelpad": 1,
        "lines.linewidth": 0.8,
        "lines.markersize": 2.5,
        "legend.fontsize": 5,
        "legend.title_fontsize": 6,
    }
    mpl.rcParams.update(params)


def myplot(
    ax=None,
    x=[0, 1],
    y=[0, 1],
    xlbl="xlbl",
    ylbl="ylbl",
    type="lin",
    s=0,
    e=-1,
    normx=1,
    normy=1,
    lbl=True,
    **kwargs,
):
    if ax is None:
        fig, ax = plt.subplots()

    if type == "lin":
        (line,) = ax.plot(x[s:e] * normx, y[s:e] * normy, **kwargs)
    elif type == "semilogy":
        (line,) = ax.semilogy(x[s:e] * normx, y[s:e] * normy, **kwargs)
    elif type == "semilogx":
        (line,) = ax.semilogx(x[s:e] * normx, y[s:e] * normy, **kwargs)
    elif type == "loglog":
        (line,) = ax.loglog(x[s:e] * normx, y[s:e] * normy, **kwargs)

    if lbl:
        ax.set_xlabel(xlbl)
        ax.set_ylabel(ylbl)

    return line




import re
import numpy as np

from buildupp.configuration import Config
from buildupp.utils import myplot

class Blend:

    @staticmethod
    def compute_surface(diameter):
        surface = 4 * np.pi * (diameter / 2) ** 2
        return surface

    @staticmethod
    def compute_volume(diameter):
        volume = 4 / 3 * np.pi * (diameter / 2) ** 3
        return volume

    def __init__(self,
                 composition: str,
                 config: Config = Config(),):

        self.config = config
        self.prop = self.config.powder_properties

        self.composition = composition
        self.pairs = self.get_pairs()

        self.density_gpcm3 = self.get_density(units="g/cm3")
        self.ssa_bet_m2pg = self.get_ssa_bet(units="m2/g")

        self.psd_dict = self.get_psd_dict()
        self.psd_diameters_mum = self.psd_dict.get("diameters_mum", 0)
        self.psd_fractions_pc = self.psd_dict.get("fractions_pc", 0)
        self.psd_cumulatives_pc = self.psd_dict.get("cumulatives_pc", 0)

        self.d10 = np.interp(
            10, self.psd_cumulatives_pc, self.psd_diameters_mum
        )  # 10% cumulative fraction
        self.d50 = np.interp(
            50, self.psd_cumulatives_pc, self.psd_diameters_mum
        )  # 50% cumulative fraction
        self.d90 = np.interp(
            90, self.psd_cumulatives_pc, self.psd_diameters_mum
        )  # 90% cumulative fraction

        self.ds = self.compute_ds()
        self.ssa_from_psd_m2pg = self.compute_ssa_from_psd(units="m2/g")

        self.rouf = self.compute_rou_over_F()

    def get_pairs(self) -> list:
        """For a given (blend) composition,
        extract pairs of powder fraction and powder ID
        Example:
            extract_pairs('85opc15ls') -> [('85', 'opc'), ('15', 'ls')]

        Args:
            composition (str): The composition of the blend, e.g '85opc15ls' or '55opc45mk'

        Returns:
            list: list of tuples (fraction (str), powder ID (str))
        """

        if self.composition == "lc355" or self.composition == "lc3":
            pairs = [("53.4", "opc"), ("1.6", "gyps"), ("15", "ls"), ("30", "mk")]
        elif self.composition == "lc3_1to1":
            pairs = [("53.4", "opc"), ("1.6", "gyps"), ("22.5", "ls"), ("22.5", "mk")]
        else:
            pairs = re.findall(r"(\d+)([a-zA-Z-]+)", self.composition)

        return pairs

    def get_psd_dict(self):
        """Compute and return the weighted PSD (Particle Size Distribution) values.

        Returns:
            dict: A dictionary containing weighted diameters, fractions, and cumulative percentages.
        """
        # Initialize weighted sums as None so we can add element-wise later
        weighted_diameter = None
        weighted_fraction = None
        weighted_cumulative = None

        # Loop through each material in the pairs
        for num, mat in self.pairs:
            # Load PSD data for the material using the Config object
            psd_df = self.config.get_psd_data(mat)
            
            # Convert fraction to float
            num = float(num) / 100  # Convert the amount to a fraction

            # Ensure the PSD dataframe has the necessary columns
            if not all(col in psd_df.columns for col in ["diameter_mum", "fraction_pc", "cumulative_pc"]):
                raise ValueError(f"Missing required PSD columns for material: {mat}")

            # Extract PSD values
            diameter = num * psd_df["diameter_mum"].values
            frac = num * psd_df["fraction_pc"].values
            cumul = num * psd_df["cumulative_pc"].values

            # Sum element-wise (initialize if first iteration)
            if weighted_diameter is None:
                weighted_diameter = diameter
                weighted_fraction = frac
                weighted_cumulative = cumul
            else:
                weighted_diameter += diameter
                weighted_fraction += frac
                weighted_cumulative += cumul

        return {
            "diameters_mum": weighted_diameter,
            "fractions_pc": weighted_fraction,
            "cumulatives_pc": weighted_cumulative,
        }


    def get_density(self, units="g/cm3") -> float:
        """Return the density in the specified unit.

        Args:
            units (str): The desired unit of density. Options: 'g/cm3', 'g/m3'.

        Returns:
            float: The density in the specified unit.
        """

        if self.composition == "lc355" or self.composition == "lc3":
            total_density = (
                0.534 * self.prop["opc"]["density_gpcm3"]
                + 0.016 * self.prop["gyps"]["density_gpcm3"]
                + 0.15 * self.prop["ls"]["density_gpcm3"]
                + 0.3 * self.prop["mk"]["density_gpcm3"]
            )

        elif self.composition == "lc3_1to1":
            total_density = (
                0.534 * self.prop["opc"]["density_gpcm3"]
                + 0.016 * self.prop["gyps"]["density_gpcm3"]
                + 0.225 * self.prop["ls"]["density_gpcm3"]
                + 0.225 * self.prop["mk"]["density_gpcm3"]
            )

        else:
            total_density = 0
            for num, mat in self.pairs:
                num = float(num)
                mat_density = self.prop.get(mat).get("density_gpcm3", 0)
                total_density += (num / 100) * float(mat_density)

        if units == "g/cm3":
            return total_density

        elif units == "g/m3":
            return total_density * 1e6  # 1 g/cm3 = 1,000,000 g/m3

        else:
            raise ValueError("Unsupported unit. Use 'g/cm3' or 'g/m3'")

    def get_ssa_bet(self, units="m2/g") -> float:
        """Return the BET SSA in the specified unit.

        Args:
            units (str): The desired unit of density.
                         Options: 'gm2/g' (default).

        Returns:
            float: The BET SSA in the specified unit.
        """

        if self.composition == "lc355" or self.composition == "lc3":
            total_ssa = (
                0.534 * self.prop["opc"]["ssa_bet_m2pg"]
                + 0.016 * self.prop["gyps"]["ssa_bet_m2pg"]
                + 0.15 * self.prop["ls"]["ssa_bet_m2pg"]
                + 0.3 * self.prop["mk"]["ssa_bet_m2pg"]
            )
        elif self.composition == "lc3_1to1":
            total_ssa = (
                0.534 * self.prop["opc"]["ssa_bet_m2pg"]
                + 0.016 * self.prop["gyps"]["ssa_bet_m2pg"]
                + 0.225 * self.prop["ls"]["ssa_bet_m2pg"]
                + 0.225 * self.prop["mk"]["ssa_bet_m2pg"]
            )
        else:
            total_ssa = 0
            for num, mat in self.pairs:
                num = float(num)
                mat_ssa = self.prop.get(mat).get("ssa_bet_m2pg", 0)
                total_ssa += (num / 100) * float(mat_ssa)

        if units == "m2/g":
            return total_ssa
        
        elif units == "cm2/g":
            return total_ssa * 1e4

        else:
            raise NotImplementedError("Specified units not implemented. Use m2/g or cm2/g.")

    def compute_ds(self, decimals=1):
        """Calculate the weighted average surface diameter based on PSD data.

        Args:
            decimals (int): The number of decimal places to round the result
                            Default is 1 decimal place.

        Returns:
            float: The weighted average surface diameter rounded to the specified number of decimal places.
        """
        s_tot = 0  # Total surface area sum
        sum_sidi = 0  # Weighted sum of surface area times diameter

        for diameter, fraction in zip(self.psd_diameters_mum, self.psd_fractions_pc):

            if np.isnan(diameter) or np.isnan(fraction):
                continue

            di = diameter  # Current diameter
            si = self.compute_surface(diameter=di)

            s_tot += si * fraction
            sum_sidi += si * di * fraction

        d_s = sum_sidi / s_tot

        return round(d_s, decimals)

    def compute_ssa_from_psd(self, units="m2/g", decimals=5):
        s_tot_m2 = 0
        v_tot_cm3 = 0

        for diameter, fraction in zip(self.psd_diameters_mum, self.psd_fractions_pc):
            if np.isnan(diameter) or np.isnan(fraction):
                continue

            di_surface = diameter * fraction * 1 / (1000 * 1000)  # mum into m
            di_volume = diameter * fraction * 1 / (1000 * 10)  # mum into cm

            si_m2 = self.compute_surface(diameter=di_surface)
            vi_cm3 = self.compute_volume(diameter=di_volume)

            s_tot_m2 += si_m2
            v_tot_cm3 += vi_cm3

        mass_g = self.get_density(units="g/cm3") * v_tot_cm3
        ssa = s_tot_m2 / mass_g

        if units == "m2/g":
            return round(ssa, decimals)

        elif units == "cm2/g":
            return round(ssa, decimals) * 1e4
        
        else:
            raise NotImplementedError("Specified units not implemented. Use m2/g or cm2/g.")

    def compute_rou_over_F(self):
        """
        Computes the ratio between roughness factor and form factor
        according to The needle model from Ouzia (CCR 2019)
        """
        s_spe = self.ssa_bet_m2pg
        M = self.density_gpcm3 * 1e6  # in m3

        # last entry contains nan for some reason
        rs = self.psd_diameters_mum[:-1] * 1e-6 / 2  # radius in m
        pofr = self.psd_fractions_pc[:-1]

        num = 0
        denom = 0

        for i in range(len(rs)):
            # Skip if rs or pofr is NaN
            if np.isnan(rs[i]) or np.isnan(pofr[i]):
                continue  # Skip this iteration

            num += rs[i] ** 3 * pofr[i]
            denom += rs[i] ** 2 * pofr[i]

        # Handle zero denominator case
        if denom == 0:
            print("Error: Denominator is zero, returning error code -1")
            return -1  # Return an error code (-1) when the denominator is zero

        rouf = round(1 / 3 * s_spe * M * num / denom, 2)
        return rouf

    def plot_psd_fractions(self, ax=None, lbl=True, type="semilogx", **kwargs):

        myplot(
            ax=ax,
            x=self.psd_diameters_mum,
            y=self.psd_fractions_pc,
            xlbl=rf"diameter [$\mu$m]",
            ylbl=rf"diff. [%]",
            lbl=lbl,
            type=type,
            **kwargs,
        )

    def plot_psd_cumulatives(self, ax=None, lbl=True, type="semilogx", **kwargs):

        myplot(
            ax=ax,
            x=self.psd_diameters_mum,
            y=self.psd_cumulatives_pc,
            xlbl=rf"diameter [$\mu$m]",
            ylbl=rf"cumul. [%]",
            lbl=lbl,
            type=type,
            **kwargs,
        )



import re
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import ConnectionPatch
from scipy.optimize import curve_fit

from pranzo import Analyzer
from buildupp.blend import Blend
from buildupp.hydration import Hydration

from buildupp.configuration import Config
from buildupp.utils import myplot


class Experiment:

    @staticmethod
    def find_nearest(array, value):
        array = np.asarray(array)
        idx = (np.abs(array - value)).argmin()
        return idx

    @staticmethod
    def convert_wc_to_decimal(s):
        match = re.search(r"wc(\d+)", s)
        if match:
            number = int(match.group(1))
            # Convert to decimal
            decimal_value = number / 100
            return decimal_value
        else:
            return None

    C3S_hydration = {
        "reactants": {"solids": {"C3S": 1}, "liquids": {"H": 3.43}},
        "products": {"solids": {"C1.67SH2.1": 1, "CH": 1.33}, "liquids": {}},
    }

    water_density_gpcm3 = 1

    def __init__(
        self,
        params: dict = None,
        config: Config = Config(),
        tstart: float = None,
        tmax: float = None,
        chemical_reaction: dict = None,
    ) -> None:

        self.config = config
        self.params = self.config.systems_measured.default[0] if params is None else params

        self.df = self.params.get("db_entry")
        self.onset_time_s = self.params.get("onset_time_s", 7500)
        self.calo_rescale = self.params.get("rescale_heat", 1)
        self.composition = self.params.get("composition", "100opc")
        self.blend = Blend(self.composition)
        self.wc = self.convert_wc_to_decimal(self.params.get("wc_string", "wc36"))
        self.mk_content_str = self.params.get("mk_content_string", "0mk")
        self.mk_content = float(re.findall(r"\d+\.?\d*", self.mk_content_str)[0])

        self.tstart = self.params.get("tstart", 3600) if tstart is None else tstart
        self.tmax = self.params.get("tmax", 4 * 3500) if tmax is None else tmax

        self.chemical_reaction = (
            self.params.get("chemical_reaction", self.C3S_hydration)
            if chemical_reaction is None
            else chemical_reaction
        )

        try:
            self.hydration = Hydration(
                chem_reaction=self.chemical_reaction, config=self.config
            )
        except FileNotFoundError:
            print(f"Cannot locate config file chemistry_database.ini")

        self.ssa_m2pg = self.blend.ssa_bet_m2pg
        self.solids_density_gpcm3 = self.blend.density_gpcm3
        self.volfrac = 1 / (
            1
            + self.wc * self.solids_density_gpcm3 / self.water_density_gpcm3
        )

        self.anlzr = Analyzer(self.df)
        self.calo = self.anlzr.calo
        self.rheo = self.anlzr.rheo.phase("p2")

        self.t = np.linspace(self.tstart, self.tmax, 10000)
        h = self.calo.interpolate("norm_heat_flow_Wpgbinder", self.t)
        H = self.calo.interpolate("norm_heat_Jpgbinder", self.t)
        self.G = self.rheo.interpolate("storage_modulus_Pa", self.t, cutoff=0.008)

        self.h, self.H = h * self.calo_rescale, H * self.calo_rescale
        self.onset_idx = self.find_nearest(self.t, self.onset_time_s)

        self.DH = self.H - self.H[self.onset_idx]
        self.G0 = self.G[self.onset_idx]
        self.Gtilde = self.G / self.G[self.onset_idx]

        self.Dm_tot_gpgbinder = self.DH * self.hydration.mu_tot_gpJ
        self.Dm_s_gpgbinder = self.DH * self.hydration.mu_s_gpJ
        self.Dm_l_gpgbinder = self.DH * self.hydration.mu_l_gpJ

        self.DV_tot_cm3pgbinder = self.DH * self.hydration.nu_tot_cm3pJ
        self.DV_s_cm3pgbinder = self.DH * self.hydration.nu_s_cm3pJ
        self.DV_l_cm3pgbinder = self.DH * self.hydration.nu_l_cm3pJ
        self.DV_s_products_cm3pgbinder = self.DH * self.hydration.nu_s_products_cm3pJ

        self.DHperSgrain_Jpm2 = self.DH / self.ssa_m2pg
        self.DVproductsperSgrain_cm3pm2 = self.DV_s_products_cm3pgbinder / self.ssa_m2pg
        self.DVproductsperVgrain = self.DV_s_cm3pgbinder * self.solids_density_gpcm3

        # compute thickness of hydration products
        self.Dhprod_mum = self.DVproductsperSgrain_cm3pm2

        # compute Dphi
        wc = self.wc
        rho_w = self.hydration.chemdat_dict["H"]["density_gpcm3"]
        rho_b = self.blend.density_gpcm3
        DVs = self.DV_s_cm3pgbinder
        DVtot = self.DV_tot_cm3pgbinder
        Dphi_nom = DVs * (1 / rho_b + wc / rho_w) - DVtot / rho_b
        Dphi_denom = (1 / rho_b + wc / rho_w) ** 2 + DVtot * (1 / rho_b + wc / rho_w)
        self.Dphi = Dphi_nom / Dphi_denom


    def plot_ht(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots heat flow [mW/g] vs. time [s]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        line = myplot(
            ax=ax,
            x=self.t * self.anlzr.nft,
            y=self.h * self.anlzr.nfh,
            xlbl=rf"time [{self.anlzr.ut}]",
            ylbl=rf"h [{self.anlzr.uh}]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_Ht(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots cumulative heat [J/g binder] vs. time [s]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        line = myplot(
            ax=ax,
            x=self.t * self.anlzr.nft,
            y=self.H * self.anlzr.nfH,
            xlbl=rf"time [{self.anlzr.ut}]",
            ylbl=rf"H [{self.anlzr.uH}]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_DHt(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots cumul heat - cumul heat onset [J/g binder] vs. time [s]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        line = myplot(
            ax=ax,
            x=self.t * self.anlzr.nft,
            y=self.DH * self.anlzr.nfH,
            xlbl=rf"time [{self.anlzr.ut}]",
            ylbl=rf"$\Delta$H [{self.anlzr.uH}]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_Gt(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots storage modulus [MPa] vs. time [s]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        line = myplot(
            ax=ax,
            x=self.t * self.anlzr.nft,
            y=self.G * self.anlzr.nfG,
            xlbl=rf"time [{self.anlzr.ut}]",
            ylbl=rf"G' [{self.anlzr.uG}]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_Gtildet(self, ax=None, lbl=True, **kwargs):
        """
        plots normalized storage modulus [-] vs. time [s]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        line = myplot(
            ax=ax,
            x=self.t * self.anlzr.nft,
            y=self.Gtilde,
            xlbl=rf"time [{self.anlzr.ut}]",
            ylbl=r"$\tilde{G}$ [-]",
            lbl=lbl,
            **kwargs,
        )

        return line

    def plot_GDH(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots storage modulus [MPa]
        vs. change in heat [J/g binder]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        line = myplot(
            ax=ax,
            x=self.DH * self.anlzr.nfH,
            y=self.G * self.anlzr.nfG,
            xlbl=rf"H-H$_0$ [{self.anlzr.uH}]",
            ylbl=rf"G' [{self.anlzr.uG}]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_GtildeDH(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots normalized storage modulus [-]
        vs. change in heat [J/g binder]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """
        G = "G"
        Gtilde = rf"$\tilde{{{G}}}$"

        line = myplot(
            ax=ax,
            x=self.DH * self.anlzr.nfH,
            y=self.Gtilde,
            xlbl=rf"$\Delta$H [{self.anlzr.uH}]",
            ylbl=rf"{Gtilde} [-]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_GDHperSgrain(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots storage modulus [MPa]
        vs. change in heat per initial grain surface [J/m2]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        line = myplot(
            ax=ax,
            x=self.DHperSgrain_Jpm2 * self.anlzr.nfH,
            y=self.G * self.anlzr.nfG,
            xlbl=rf"H-H$_0$ [J/m$^2$]",
            ylbl=rf"G' [{self.anlzr.uG}]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_GtildeDHperSgrain(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots normalized storage modulus [-]
        vs. change in heat per initial grain surface [J/m2]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """
        G = "G"
        Gtilde = rf"$\tilde{{{G}}}$"

        line = myplot(
            ax=ax,
            x=self.DHperSgrain_Jpm2,
            y=self.Gtilde,
            xlbl=rf"$\Delta$H [J/m$^2$]",
            ylbl=rf"{Gtilde} [-]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_GDVproducts(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots storage modulus [MPa]
        vs. change in volume  of products per initial grain surface [cm3/ g binder]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        line = myplot(
            ax=ax,
            x=self.DV_products_cm3pgbinder,
            y=self.G * self.anlzr.nfG,
            xlbl=r"V$^{\mathrm{prod}}$-V$_0$$^\mathrm{prod}$ [cm$^3$/g]",
            ylbl=rf"G' [{self.anlzr.uG}]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_GtildeDVproducts(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots normalized storage modulus [-]
        vs. change in volume of products per unit binder [cm3/g binder]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """
        G = "G"
        Gtilde = rf"$\tilde{{{G}}}$"

        line = myplot(
            ax=ax,
            x=self.DV_products_cm3pgbinder,
            y=self.Gtilde,
            xlbl=r"V$^{\mathrm{prod}}$-V$_0$$^\mathrm{prod}$ [cm$^3$/g]",
            ylbl=rf"{Gtilde} [-]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_GDVproductsperSgrain(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots storage modulus [MPa]
        vs. change in volume per initial grain surface [cm3/m2]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        line = myplot(
            ax=ax,
            x=self.DVproductsperSgrain_cm3pm2,
            y=self.G * self.anlzr.nfG,
            xlbl=r"(V$^{\mathrm{prod}}$-V$_0$$^\mathrm{prod}$) / S$_{\mathrm{grain}}$ [cm$^3$/m$^2$]",
            ylbl=rf"G' [{self.anlzr.uG}]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_GtildeDVproductsperSgrain(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots normalized storage modulus [-]
        vs. change in volume of products per initial grain surface [cm3/m2]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """
        G = "G"
        Gtilde = rf"$\tilde{{{G}}}$"

        line = myplot(
            ax=ax,
            x=self.DVproductsperSgrain_cm3pm2,
            y=self.Gtilde,
            xlbl=r"(V$^{\mathrm{prod}}$-V$_0$$^\mathrm{prod}$) / S$_{\mathrm{grain}}$ [cm$^3$/m$^2$]",
            ylbl=rf"{Gtilde} [-]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_GDVproductsperVgrain(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots storage modulus [MPa]
        vs. change in volume per initial grain volume

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        line = myplot(
            ax=ax,
            x=self.DVproductsperVgrain,
            y=self.G * self.anlzr.nfG,
            xlbl=r"(V$^{\mathrm{prod}}$-V$_0$$^\mathrm{prod}$) / V$_{\mathrm{grain}}$ [-]",
            ylbl=rf"G' [{self.anlzr.uG}]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_GtildeDVproductsperVgrain(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots normalized storage modulus [-]
        vs. change in volume of products per initial grain volume

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        G = "G"
        Gtilde = rf"$\tilde{{{G}}}$"

        line = myplot(
            ax=ax,
            x=self.DVproductsperVgrain,
            y=self.Gtilde,
            xlbl=r"(V$^{\mathrm{prod}}$-V$_0$$^\mathrm{prod}$) / V$_{\mathrm{grain}}$ [-]",
            ylbl=rf"{Gtilde} [-]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line
    
    def plot_GDhprod_mum(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots storage modulus [MPa]
        vs. change in thickness in products thickness (h) [mum]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        line = myplot(
            ax=ax,
            x=self.Dhprod_mum,
            y=self.G * self.anlzr.nfG,
            xlbl=r"change in products thickness h-h$_0$[$\mu$m]",
            ylbl=rf"G' [{self.anlzr.uG}]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_GtildeDhprod_mum(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots normalized storage modulus [-]
        vs. change in thickness in products thickness (h) [mum]

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """
        G = "G"
        Gtilde = rf"$\tilde{{{G}}}$"

        line = myplot(
            ax=ax,
            x=self.DVproductsperSgrain_cm3pm2,
            y=self.Gtilde,
            xlbl=r"change in products thickness h-h$_0$[$\mu$m]",
            ylbl=rf"{Gtilde} [-]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def get_fit_params(self, fit, x="DH", y="G", xx0_bound=0, period=None):
        """
        Fit parameters for specified x and y data using a given fitting function.

        This method retrieves x and y values based on the specified keys,
        finds the nearest index for a given boundary, and performs curve fitting
        using the provided fitting function. The fitting can be performed on
        different segments of the data based on the specified period.
        induction = up to the boundary
        acceleration = from the boundary onwards

        Parameters:
            fit : callable
                The fitting function to use for curve fitting. It should take
                x values as the first argument and any additional parameters
                as subsequent arguments.

            x : str, optional
                The key for the x values to be used for fitting.
                Default is "DH". Options include:
                - "DH"
                - "DHperSgrain"
                - "G"
                - "Gtilde"
                - "DV"
                - "DVproductsperSgrain"
                - "Dhprod_mum"

            y : str, optional
                The key for the y values to be used for fitting.
                Default is "G". Options are similar to x.

            xx0_bound : float, optional
                The boundary value for segmenting the data. The nearest index
                to this value will be found in the x values. Default is 0.

            period : str or None, optional
                Specifies the segment of the data to fit:
                - None: Fit on the entire dataset.
                - "induct": Fit on the data up to the boundary.
                - "accel": Fit on the data after the boundary.

        Returns:
            np.ndarray
                The optimal parameters for the fit obtained from the curve fitting.

        Raises:
            ValueError: If the specified x or y keys do not exist in the conversion_dict.
        """

        conversion_dict = {
            "DH": self.DH,
            "DHperSgrain": self.DHperSgrain_Jpm2,
            "G": self.G,
            "Gtilde": self.Gtilde,
            "DV": self.DV_s_products_cm3pgbinder,
            "DVproductsperSgrain": self.DVproductsperSgrain_cm3pm2,
            "Dhprod_mum": self.Dhprod_mum,
        }

        x_values = conversion_dict.get(x, None)
        y_values = conversion_dict.get(y, None)

        idx_bound = self.find_nearest(array=x_values, value=xx0_bound)

        if period == "induct":
            x_fit = x_values[:idx_bound]
            y_fit = y_values[:idx_bound]
        elif period == "accel":
            x_fit = x_values[idx_bound:]
            y_fit = y_values[idx_bound:]
        else:  # period is None or any other value
            x_fit = x_values
            y_fit = y_values

        p, _ = curve_fit(fit, x_fit, y_fit)

        return p

    def plot_Dphi_t(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots change in solid volme fraction over time

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        line = myplot(
            ax=ax,
            x=self.t * self.anlzr.nft,
            y=self.Dphi,
            xlbl=rf"time [{self.anlzr.ut}]",
            ylbl=r"$\Delta \phi$ [-]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def plot_Dphi_DH(self, ax=None, lbl=True, s=0, e=-1, **kwargs):
        """
        plots change in solid volme fraction against
        changes in cumulative heat

        Args:
            ax: define custom axis
            lbl: displays axes labels, default is True
            s: start index, default is 0
            e: end index, default is -1
        """

        line = myplot(
            ax=ax,
            x=self.DH * self.anlzr.nfH,
            y=self.Dphi,
            xlbl=rf"$\Delta$H [{self.anlzr.uH}]",
            ylbl=r"$\Delta \phi$ [-]",
            lbl=lbl,
            s=s,
            e=e,
            **kwargs,
        )

        return line

    def draw_onset(
        self,
        ax_ht,
        ax_HH0t,
        ax_Gt,
        cross_color="k",
        cross_fntsz=5,
        cross_alpha=0.7,
        line_color="0.3",
        line_style="-",
        line_width=0.4,
        line_alpha=0.4,
        line_prolong=1,
    ):

        # add cross at onset in h-t
        ax_ht.text(
            self.t[self.onset_idx] * self.anlzr.nft,
            self.h[self.onset_idx] * self.anlzr.nfh,
            "x",
            color=cross_color,
            alpha=cross_alpha,
            fontsize=cross_fntsz,
            horizontalalignment="center",
            verticalalignment="center",
        )

        ax_HH0t.text(
            self.t[self.onset_idx] * self.anlzr.nft,
            self.DH[self.onset_idx],
            "x",
            color=cross_color,
            alpha=cross_alpha,
            fontsize=cross_fntsz,
            horizontalalignment="center",
            verticalalignment="center",
        )

        ax_Gt.text(
            self.t[self.onset_idx] * self.anlzr.nft,
            self.G[self.onset_idx] * self.anlzr.nfG,
            "x",
            color=cross_color,
            alpha=cross_alpha,
            fontsize=cross_fntsz,
            horizontalalignment="center",
            verticalalignment="center",
        )

        #### vlines through time evolutions
        xy_ht = (
            self.t[self.onset_idx] * self.anlzr.nft,
            self.h[self.onset_idx] * self.anlzr.nfh,
        )
        xy_Gt = (
            self.t[self.onset_idx] * self.anlzr.nft,
            self.G[self.onset_idx] * self.anlzr.nfG,
        )
        con = ConnectionPatch(
            xyA=xy_ht,
            xyB=xy_Gt,
            coordsA="data",
            coordsB="data",
            axesA=ax_ht,
            axesB=ax_Gt,
            color=line_color,
            lw=line_width,
            linestyle=line_style,
            alpha=line_alpha,
        )
        ax_Gt.add_artist(con)

        #### longer lines for "onset" text
        xy_ht_1 = (
            self.t[self.onset_idx] * self.anlzr.nft,
            self.h[self.onset_idx] * self.anlzr.nfh,
        )
        xy_ht_2 = (self.t[self.onset_idx] * self.anlzr.nft, line_prolong)
        bmp = ConnectionPatch(
            xyA=xy_ht_1,
            xyB=xy_ht_2,
            coordsA="data",
            coordsB="data",
            axesA=ax_ht,
            axesB=ax_ht,
            color=line_color,
            lw=line_width,
            linestyle=line_style,
            alpha=line_alpha,
        )
        ax_ht.add_artist(bmp)

    def text_onset_H0_G0(
        self,
        ax_ht,
        ax_HH0t,
        ax_Gt,
        text_color="k",
        text_fontsize=6,
        text_alpha=0.8,
        onset_xy=[2.1, 1],
        H0_xy=[2.1, 1],
        G0_xy=[2.1, 1],
    ):

        ## onset, H0 and G0 texts
        ax_ht.text(
            onset_xy[0],
            onset_xy[1],
            "onset",
            color=text_color,
            fontsize=text_fontsize,
            alpha=text_alpha,
        )

        ax_HH0t.text(
            H0_xy[0],
            H0_xy[1],
            r"H$_{\mathrm{0}}$",
            color=text_color,
            fontsize=text_fontsize,
            alpha=text_alpha,
        )

        ax_Gt.text(
            G0_xy[0],
            G0_xy[1],
            r"G$_{\mathrm{0}}'$",
            color=text_color,
            fontsize=text_fontsize,
            alpha=text_alpha,
        )



import pytest
import numpy as np
from buildupp.blend import Blend


def test_compute_surface():
    assert Blend.compute_surface(1) == np.pi

def test_compute_volume():
    assert Blend.compute_volume(1) == np.pi / 6

def test_get_pairs():
    blend_1 = Blend('85pc15ls')
    assert blend_1.get_pairs() == [('85', 'pc'), ('15', 'ls')]

    blend_2 = Blend('85opc15ls')
    assert blend_2.get_pairs() == [('85', 'opc'), ('15', 'ls')]

    blend_3 = Blend('55opc_45lsfine')
    assert blend_3.get_pairs() == [('55', 'opc'), ('45', 'lsfine')]

# this one is important, make dummy psd files with like perfect bidisperse
# the d50 and d90 likely fail
def test_get_psd_dict():
    pass

def test_get_density():
    opc = Blend('100opc')
    assert opc.get_density(units='g/cm3') == 3.15
    assert opc.get_density(units='g/m3') == 3.15 * 1e6

    ls = Blend('100ls')
    assert ls.get_density(units='g/cm3') == 2.7
    assert ls.get_density(units='g/m3') == 2.7 * 1e6

    opc_ls_5545 = Blend('55opc45ls')
    assert opc_ls_5545.get_density(units='g/cm3') == 0.55 * 3.15 + 0.45 * 2.7
    assert opc_ls_5545.get_density(units='g/m3') == (0.55 * 3.15 + 0.45 * 2.7) * 1e6

def test_get_ssa_bet():
    opc = Blend('100opc')
    assert opc.get_ssa_bet(units='m2/g') == 1.09
    assert opc.get_ssa_bet(units='cm2/g') == 1.09 * 1e4

    ls = Blend('100ls')
    assert ls.get_ssa_bet(units='m2/g') == 0.56
    assert ls.get_ssa_bet(units='cm2/g') == 0.56 * 1e4

    opc_ls_5545 = Blend('55opc45ls')
    assert opc_ls_5545.get_ssa_bet(units='m2/g') == 0.55 * 1.09 + 0.45 * 0.56
    assert opc_ls_5545.get_ssa_bet(units='cm2/g') == (0.55 * 1.09 + 0.45 * 0.56) * 1e4

def test_compute_ds():
    pass

def test_compute_ssa_from_psd():
    pass

def test_compute_rou_over_F():
    pass



from buildupp.configuration import Config

class Compound:
    """
    Compound is a compound of a chemical reaction.
    Allows to access data in the chemistry database (ini format).

    The database format is as follows:
    [compound_name]
    molar_mass_gpmol = value
    density_gpcm3 = value
    formation_enthalpy_Jpmol = value
    ref = source

    Example INI structure:
    [CH]
    molar_mass_gpmol = 74
    density_gpcm3 = 2.24
    formation_enthalpy_Jpmol = -985000
    ref = matschei_ccr_2007

    The database can be accessed using self.chemdat_dict.
    Example: self.chemdat_dict['CH']['density_gpcm3']
    """

    def __init__(self, config: Config = Config()) -> None:

        self.chemdat_dict = config.chemdat_dict

    def formation_enthalpy_Jpmol(self, compound):
        """Get the formation enthalpy of the compound (in J/mol)."""
        return self.chemdat_dict[compound]['formation_enthalpy_jpmol']

    def molar_mass_gpmol(self, compound):
        """Get the molar mass of the compound (in g/mol)."""
        return self.chemdat_dict[compound]['molar_mass_gpmol']

    def density_gpcm3(self, compound):
        """Get the density of the compound (in g/cm³)."""
        return self.chemdat_dict[compound]['density_gpcm3']


class Hydration(Compound):
    """
    Hydration reaction is based on a dictionary containing
    the chemical equation with the stoichiometric coefficients
    and on a INI file containing the enthalpies of formation of the
    reactants and products.

    Args:
      - chem_reaction: dict

        example dict for C3S hydration:

        C3S_hydration = {'reactants':{'solids':{'C3S':1},
                                    'liquids':{'H':3.43}},
                        'products':{'solids':{'C1.67SH2.1':1, 'CH':1.33},
                                    'liquids':{}}}

      - db: ini file containing the enthalpies of formation
        For the format of the ini file, see class Compound
    """

    def __init__(self, chem_reaction: dict, config: Config = Config()) -> None:
        super().__init__(config=config)


        self.chemical_reaction = chem_reaction

        ## returns dict, e.g. {'C3S':1}
        self.reactants_solid = chem_reaction["reactants"]["solids"]
        self.reactants_liquid = chem_reaction["reactants"]["liquids"]
        self.products_solid = chem_reaction["products"]["solids"]
        self.products_liquid = chem_reaction["products"]["liquids"]

        self.DH0reaction_Jpmol = self.compute_DH0reaction_Jpmol()

        self.mtotchange_gpmol = self.compute_mtotchange_gpmol()
        self.mschange_gpmol = self.compute_mschange_gpmol()
        self.mlchange_gpmol = self.compute_mlchange_gpmol()
        self.mschange_products_gpmol = self.compute_mschange_products_gpmol()

        self.Vtot_change_cm3pmol = self.compute_Vtot_change_cm3pmol()
        self.Vs_change_cm3pmol = self.compute_Vs_change_cm3pmol()
        self.Vl_change_cm3pmol = self.compute_Vl_change_cm3pmol()
        self.Vs_change_products_cm3pmol = self.compute_Vs_change_products_cm3pmol()

        self.mu_tot_gpJ = self.compute_mu_tot_gpJ()
        self.mu_s_gpJ = self.compute_mu_s_gpJ()
        self.mu_l_gpJ = self.compute_mu_l_gpJ()
        self.mu_s_products_gpJ = self.compute_mu_s_products_gpJ()

        self.nu_tot_cm3pJ = self.compute_nu_tot_cm3pJ()
        self.nu_s_cm3pJ = self.compute_nu_s_cm3pJ()
        self.nu_l_cm3pJ = self.compute_nu_l_cm3pJ()
        self.nu_s_products_cm3pJ = self.compute_nu_s_products_cm3pJ()

    def compute_DH0reaction_Jpmol(self):
        """Compute the enthalpy of reaction Delta H^0_{reaction} from Hess's law

        Computes the heat generated (sign depends if endo- or exothremic)
        by the reaction for the formation of 1 mol of product.

        For the default C3S hydration, it computes it for the
        formation of 1 mol of C-S-H.
        """

        # DH_reaction = SUM(n_prod DH_prod) - SUM(n_reactants DH_reactants)
        DH_reaction = 0

        # products_solid[compound] gives the number of moles of compound in the products that are solids
        for compound in self.products_solid:
            DH_reaction += self.products_solid[
                compound
            ] * self.formation_enthalpy_Jpmol(compound)
        for compound in self.products_liquid:
            DH_reaction += self.products_liquid[
                compound
            ] * self.formation_enthalpy_Jpmol(compound)
        for compound in self.reactants_solid:
            DH_reaction -= self.reactants_solid[
                compound
            ] * self.formation_enthalpy_Jpmol(compound)
        for compound in self.reactants_liquid:
            DH_reaction -= self.reactants_liquid[
                compound
            ] * self.formation_enthalpy_Jpmol(compound)

        return DH_reaction

    def compute_mtotchange_gpmol(self):
            """Compute the total mass change for 1 mol of product.

            For the default C3S hydration, it computes it for the
            formation of 1 mol of C-S-H.
            """

            mtot_change = 0

            for compound in self.products_solid:
                mtot_change += self.products_solid[compound] * self.molar_mass_gpmol(compound)
            
            for compound in self.products_liquid:
                mtot_change += self.products_liquid[compound] * self.molar_mass_gpmol(compound)

            for compound in self.reactants_solid:
                mtot_change -= self.reactants_solid[compound] * self.molar_mass_gpmol(
                    compound
                )

            for compound in self.reactants_liquid:
                mtot_change -= self.reactants_liquid[compound] * self.molar_mass_gpmol(
                    compound
                )

            return mtot_change
    
    def compute_mschange_gpmol(self):
        """Compute the mass change of solids for 1 mol of product.

        For the default C3S hydration, it computes it for the
        formation of 1 mol of C-S-H.
        """

        ms_change = 0

        for compound in self.products_solid:
            ms_change += self.products_solid[compound] * self.molar_mass_gpmol(compound)
        for compound in self.reactants_solid:
            ms_change -= self.reactants_solid[compound] * self.molar_mass_gpmol(
                compound
            )

        return ms_change
    
    def compute_mlchange_gpmol(self):
        """Compute the mass change of liquids for 1 mol of product.

        For the default C3S hydration, it computes it for the
        formation of 1 mol of C-S-H.
        """

        ml_change = 0

        for compound in self.products_liquid:
            ml_change += self.products_liquid[compound] * self.molar_mass_gpmol(compound)
        for compound in self.reactants_liquid:
            ml_change -= self.reactants_liquid[compound] * self.molar_mass_gpmol(
                compound
            )

        return ml_change
    
    def compute_mschange_products_gpmol(self):
        """Compute the mass change of products for 1 mol of product.

        For the default C3S hydration, it computes it for the
        formation of 1 mol of C-S-H.
        """

        ms_change_products = 0

        for compound in self.products_liquid:
            ms_change_products += self.products_solid[compound] * self.molar_mass_gpmol(compound)

        return ms_change_products

    def compute_mu_tot_gpJ(self):
        """Compute mu_s, giving the total mass change of solids
        (in g) for 1 joule of reaction.
        """

        mu_tot = self.mtotchange_gpmol / abs(self.DH0reaction_Jpmol)
        return mu_tot
    
    def compute_mu_s_gpJ(self):
        """Compute mu_s, giving the total mass change of solids
        (in g) for 1 joule of reaction.
        """

        mu_s = self.mschange_gpmol / abs(self.DH0reaction_Jpmol)
        return mu_s
    
    def compute_mu_l_gpJ(self):
        """Compute mu_l, giving the mass change of liquids
        (in g) for 1 joule of reaction.
        """

        mu_l = self.mlchange_gpmol / abs(self.DH0reaction_Jpmol)
        return mu_l
    
    def compute_mu_s_products_gpJ(self):
        """Compute mu_s_products, giving the mass change of solid products
        (in g) for 1 joule of reaction.
        """

        mu_s_products = self.mschange_gpmol / abs(self.DH0reaction_Jpmol)
        return mu_s_products
    
    def compute_Vtot_change_cm3pmol(self):
        """Compute the change in total volume
        (in cm3) for 1 mol of product.

        For the default C3S hydration, it computes it for the
        formation of 1 mol of C-S-H.
        """
        Vtot_change = 0

        for compound in self.products_solid:
            Vtot_change += (
                self.products_solid[compound]
                * self.molar_mass_gpmol(compound)
                / self.density_gpcm3(compound)
            )
        
        for compound in self.products_liquid:
            Vtot_change += (
                self.products_liquid[compound]
                * self.molar_mass_gpmol(compound)
                / self.density_gpcm3(compound)
            )

        for compound in self.reactants_solid:
            Vtot_change -= (
                self.reactants_solid[compound]
                * self.molar_mass_gpmol(compound)
                / self.density_gpcm3(compound)
            )

        for compound in self.reactants_liquid:
            Vtot_change -= (
                self.reactants_liquid[compound]
                * self.molar_mass_gpmol(compound)
                / self.density_gpcm3(compound)
            )

        return Vtot_change

    def compute_Vs_change_cm3pmol(self):
        """Compute the total change in volume of solids
        (in cm3) for 1 mol of product.

        For the default C3S hydration, it computes it for the
        formation of 1 mol of C-S-H.
        """
        Vs_change = 0

        for compound in self.products_solid:
            Vs_change += (
                self.products_solid[compound]
                * self.molar_mass_gpmol(compound)
                / self.density_gpcm3(compound)
            )

        for compound in self.reactants_solid:
            Vs_change -= (
                self.reactants_solid[compound]
                * self.molar_mass_gpmol(compound)
                / self.density_gpcm3(compound)
            )

        return Vs_change
    
    def compute_Vl_change_cm3pmol(self):
        """Compute the total change in volume of liquids
        (in cm3) for 1 mol of product.

        For the default C3S hydration, it computes it for the
        formation of 1 mol of C-S-H.
        """
        Vl_change = 0

        for compound in self.products_liquid:
            Vl_change += (
                self.products_liquid[compound]
                * self.molar_mass_gpmol(compound)
                / self.density_gpcm3(compound)
            )

        for compound in self.reactants_liquid:
            Vl_change -= (
                self.reactants_liquid[compound]
                * self.molar_mass_gpmol(compound)
                / self.density_gpcm3(compound)
            )

        return Vl_change

    def compute_Vs_change_products_cm3pmol(self):
        """Compute the change in volume of solid products (in cm3)
        for 1 mol of product.

        For the default C3S hydration, it computes it for the
        formation of 1 mol of C-S-H.
        """
        Vs_change_products = 0

        for compound in self.products_solid:
            Vs_change_products += (
                self.products_solid[compound]
                * self.molar_mass_gpmol(compound)
                / self.density_gpcm3(compound)
            )

        return Vs_change_products

    def compute_nu_tot_cm3pJ(self):
        """Compute nu, giving the total volume change (in cm3)
        for 1 joule of reaction.
        """

        nu_tot = self.Vtot_change_cm3pmol / abs(self.DH0reaction_Jpmol)
        return nu_tot
    
    def compute_nu_s_cm3pJ(self):
        """Compute nu, giving the volume change of all solids (in cm3)
        for 1 joule of reaction.
        """

        nu_s = self.Vs_change_cm3pmol / abs(self.DH0reaction_Jpmol)
        return nu_s
    
    def compute_nu_l_cm3pJ(self):
        """Compute nu, giving the volume change of liquids (in cm3)
        for 1 joule of reaction.
        """

        nu_l = self.Vl_change_cm3pmol / abs(self.DH0reaction_Jpmol)
        return nu_l
    
    def compute_nu_s_products_cm3pJ(self):
        """Compute nu_s_products, giving the volume change of solid
        products (in cm3) for 1 joule of reaction.
        """

        nu_s_products = self.compute_Vs_change_products_cm3pmol() / abs(self.DH0reaction_Jpmol)
        return nu_s_products



import pytest
import numpy as np
from buildupp.hydration import Compound, Hydration

# compare to values in chemistry_database.ini
def test_formation_enthalpy_Jpmol():
    cmpnd = Compound()
    assert cmpnd.formation_enthalpy_Jpmol('C3S') == -2931000
    assert cmpnd.molar_mass_gpmol('C3S') == 228
    assert cmpnd.density_gpcm3('C3S') == 3.12

    assert cmpnd.formation_enthalpy_Jpmol('H') == -286000


# define C3S hydration
C3S_hydration = {'reactants':{'solids':{'C3S':1},
                                    'liquids':{'H':3.43}},
                        'products':{'solids':{'C1.67SH2.1':1, 'CH':1.33},
                                    'liquids':{}}}
c3s_hyd = Hydration(C3S_hydration)

# compare to values computed in master thesis report
def test_compute_DH0reaction_Jpmol():
    assert c3s_hyd.compute_DH0reaction_Jpmol() == pytest.approx(-121100, abs=50)

def test_compute_mschange_gpmol():
    assert c3s_hyd.compute_mschange_gpmol() == pytest.approx(61.7, abs=0.5)

def test_compute_mu_s_gpJ():
    assert c3s_hyd.compute_mu_s_gpJ() == pytest.approx(0.00051, abs=0.00005)

def test_compute_Vs_change_cm3pmol():
    assert c3s_hyd.compute_Vs_change_cm3pmol() == pytest.approx(48.9, abs=0.5)

def test_compute_nu_s_cm3pJ():
    assert c3s_hyd.compute_nu_s_cm3pJ() == pytest.approx(0.00041, abs=0.00005)


import configparser
import os
import sys
import importlib
import pandas as pd

import buildupp


class UserNotFoundError(AttributeError):
    pass


class Config:

    def __init__(self, config_dir: str = os.path.join(buildupp.PROJECT_ROOT, "config")):

        self.config_dir = config_dir
        self.paths_file = os.path.join(self.config_dir, "paths.py")

        # Dynamically load paths.py file
        self.data_root_directory, self.desktop_directory = self.load_paths()

        # Dynamically load .py files
        self.user_functions = self.load_user_functions()
        self.systems_measured = self.load_systems_measured()
        self.colors = self.load_colors()

        # load chemistry data
        self.chemistry_database = config_dir + "/chemistry_database.ini"
        self.chemdat_dict = self.load_chemistry_database()

        # load powder properties
        self.powder_properties = self.load_powder_properties()

    def load_paths(self):
        """Dynamically loads paths.py and retrieves the data root and desktop directories for the current user."""

        if not os.path.isfile(self.paths_file):
            raise FileNotFoundError(f"File {self.paths_file} not found.")

        # Dynamically load the paths.py file
        spec = importlib.util.spec_from_file_location("paths", self.paths_file)
        paths_module = importlib.util.module_from_spec(spec)
        sys.modules["paths"] = paths_module
        spec.loader.exec_module(paths_module)

        data_root = paths_module.data_root_directory
        desktop = paths_module.desktop_directory

        return data_root, desktop

    def load_colors(self):
        """Dynamically loads the colors.py file from the config directory."""
        colors_file = os.path.join(self.config_dir, "colors.py")

        if not os.path.isfile(colors_file):
            raise FileNotFoundError(f"File {colors_file} not found.")

        # Dynamically load the colors.py file
        spec = importlib.util.spec_from_file_location("colors", colors_file)
        colors_module = importlib.util.module_from_spec(spec)
        sys.modules["systems_measured"] = colors_module
        spec.loader.exec_module(colors_module)

        return colors_module

    def load_systems_measured(self):
        """Dynamically loads the systems_measured.py file from the config directory."""
        systems_measured_file = os.path.join(self.config_dir, "systems_measured.py")

        if not os.path.isfile(systems_measured_file):
            raise FileNotFoundError(f"File {systems_measured_file} not found.")

        # Dynamically load the systems_measured.py file
        spec = importlib.util.spec_from_file_location(
            "systems_measured", systems_measured_file
        )
        systems_measured_module = importlib.util.module_from_spec(spec)
        sys.modules["systems_measured"] = systems_measured_module
        spec.loader.exec_module(systems_measured_module)

        return systems_measured_module

    def load_chemistry_database(self):
        """Loads the INI database into a dictionary."""
        config = configparser.ConfigParser()
        config.read(self.chemistry_database)

        chemdat_dict = {}
        for section in config.sections():
            compound_data = {}
            for key in config[section]:
                try:
                    # Try to convert values to float if possible
                    compound_data[key] = (
                        float(config[section][key])
                        if "." in config[section][key]
                        else int(config[section][key])
                    )
                except ValueError:
                    # If it's not a number, store as a string (e.g., reference)
                    compound_data[key] = config[section][key]
            chemdat_dict[section] = compound_data

        return chemdat_dict  # This returns the dictionary so it can be assigned to self.chemdat_dict

    def load_powder_properties(self):
        """
        Load powder properties from a Python file (physical_properties.py) in the config directory.
        Returns a dictionary of powder properties, where each powder is a key.
        """

        properties_file = os.path.join(self.config_dir, "physical_properties.py")

        if not os.path.isfile(properties_file):
            raise FileNotFoundError(
                f"{properties_file} not found in the config directory."
            )

        # Dynamically load the Python file as a module
        spec = importlib.util.spec_from_file_location(
            "physical_properties", properties_file
        )
        physicalties_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(physicalties_module)

        physical_properties = getattr(physicalties_module, "physical_properties", {})

        # Check if the dictionary exists in the file
        if not physical_properties:
            raise ValueError(
                "No properties dictionary found in the physical_properties.py file."
            )

        return physical_properties

    def load_user_functions(self):
        """Dynamically loads user-defined functions for reading PSD data."""
        user_functions_file = os.path.join(self.config_dir, "user_functions.py")

        if not os.path.isfile(user_functions_file):
            raise FileNotFoundError(f"File {user_functions_file} not found.")

        # Dynamically load the user functions file (user_functions.py)
        spec = importlib.util.spec_from_file_location(
            "user_functions", user_functions_file
        )
        user_functions_module = importlib.util.module_from_spec(spec)
        sys.modules["user_functions"] = user_functions_module
        spec.loader.exec_module(user_functions_module)

        return user_functions_module

    def get_psd_data(self, material: str):
        """Retrieve the PSD data as a pandas DataFrame for a given material."""
        if material not in self.powder_properties:
            raise ValueError(f"Material {material} not found in powder properties.")

        # Fetch the PSD file path and reader function name from powder_properties
        psd_path = self.powder_properties[material].get("psd_file")
        reader_function_name = self.powder_properties[material].get(
            "psd_reader_function"
        )

        if not psd_path or not reader_function_name:
            raise ValueError(
                f"Missing PSD path or reader function for material {material}."
            )

        # Get the reader function dynamically from the user_functions module
        reader_function = getattr(self.user_functions, reader_function_name, None)

        if not reader_function:
            raise AttributeError(
                f"Reader function {reader_function_name} not found in user_functions.py."
            )

        psd_df = reader_function(psd_path)

        if not isinstance(psd_df, pd.DataFrame):
            raise TypeError(
                f"Reader function {reader_function_name} did not return a pandas DataFrame."
            )

        return psd_df
